--[[
Copyright (c) 2011 WangXH <kurapica.igas@gmail.com>

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
--]]

-- Class System
-- Author: kurapica.igas@gmail.com
-- Create Date : 2011/02/01
-- ChangeLog   :

------------------------------------------------------------------------
-- Class system is used to provide a object-oriented system in lua.
-- With this system, you can created a class like
--
-- namespace "System"				-- define the namespace
--
-- class "MyClass"						-- declare starting to define the class
--		inherit "Object"					-- declare the class is inherited from System.Object
--
--		script "OnNameChanged"	-- declare the class have a script named "OnNameChanged"
--
--		function Print(self)				-- the global functions will be treated as the class's methodes, self means the object
--			print(self._Name)
--		end
--
--		property "Name" {				-- declare the class have a property named "Name"
--			Get = function(self)		-- the get method for property "Name"
--				return self._Name
--			end,
--			Set = function(self, value)	-- the set method for property "Name"
--				self._Name = value
--				self:OnNameChanged(value)	-- raise the "OnNameChanged" script to trigger it's handler functions.
--			end,
--			Type = String,				-- the property "Name"'s type, so when you assign a value to Name, it should be checked.
--		}
--
--		function MyClass(name)		-- the function with same name of the class is treated as the Constructor of the class
--			local obj = {_Name = name}	-- the object should be a table
--			return obj
--		end
--	endclass "MyClass"				-- declare the definition of the class is over.
--
--	Using MyClass:
--
--	myObj = MyClass("Test")
--
--	myObj:Print()						-- print out : Test
--
--	function myObj:OnNameChanged(name)	-- define the script handler for 'OnNameChanged'
--		print("The Name is changed to "..name)
--	end
--
--	myObj.Name = "Hello"			-- print out : The Name is changed to Hello
------------------------------------------------------------------------

local version = 1

------------------------------------------------------
-- Version Check & Class Environment
------------------------------------------------------
do
	local _Meta = getmetatable(IGAS)

	-- Version Check
	if _Meta._IGAS_Class_Version and _Meta._IGAS_Class_Version >= version then
		return
	end

	_Meta._IGAS_Class_Version = version

	-- Class Environment
	_Meta._Class_ENV = _Meta._Class_ENV or {}
	_Meta._Class_ENV._Class_KeyWords = _Meta._Class_KeyWords

	setmetatable(_Meta._Class_ENV, {
		__index = function(self,  key)
			if type(key) == "string" and key ~= "_G" and key:find("^_") then
				return
			end

			if _G[key] then
				rawset(self, key, _G[key])
				return rawget(self, key)
			end
		end,
	})

	setfenv(1, _Meta._Class_ENV)

	-- Scripts
	strtrim = strtrim or function(s)
	  return s and (s:gsub("^%s*(.-)%s*$", "%1")) or ""
	end

	wipe = wipe or function(t)
		for k in pairs(t) do
			t[k] = nil
		end
	end

	geterrorhandler = geterrorhandler or function()
		return print
	end

	errorhandler = errorhandler or function(err)
		return geterrorhandler()(err)
	end

	tinsert = tinsert or table.insert
	tremove = tremove or table.remove
	unpack = unpack
	pcall = pcall

	TYPE_CLASS = "Class"
	TYPE_ENUM = "Enum"
	TYPE_STRUCT = "Struct"

	TYPE_NAMESPACE = "IGAS.NameSpace"
end

------------------------------------------------------
-- NameSpace
------------------------------------------------------
do
	_NameSpace = _NameSpace or newproxy(true)

	_NSInfo = _NSInfo or setmetatable({}, {
		__index = function(self, key)
			if not IsNameSpace(key) then
				return
			end

			self[key] = {
				Owner = key,
			}

			return rawget(self, key)
		end,
	})

	-- metatable for class
	_MetaNS = _MetaNS or getmetatable(_NameSpace)
	do
		_MetaNS.__call = function(self, ...)
			local info = _NSInfo[self]

			if info.Type == TYPE_CLASS then
				return Class2Obj(self, ...)
			elseif info.Type == TYPE_STRUCT then
				return Struct2Obj(self, ...)
			end

			error(("%s can't be used as a constructor."):format(info.Name), 2)
		end

		_MetaNS.__index = function(self, key)
			local info = _NSInfo[self]

			if info.Type == TYPE_STRUCT then
				if key == "Validate" then
					return info.WrapValidate
				end
				return info.SubNS and info.SubNS[key]
			elseif info.Type == TYPE_CLASS then
				if info.SubNS and info.SubNS[key] then
					return info.SubNS[key]
				elseif type(key) == "string" and (not key:find("^_")) and info.Cache4Method and info.Cache4Method[key] then
					return info.Cache4Method[key]
				elseif _KeyMeta[key] ~= nil then
					if _KeyMeta[key] then
						return info.MetaTable[key]
					else
						return info.MetaTable["_"..key]
					end
				end
			elseif info.Type == TYPE_ENUM then
				return type(key) == "string" and info.Enum[key:upper()] or nil
			else
				return info.SubNS and info.SubNS[key]
			end
		end

		_MetaNS.__newindex = function(self, key, value)
			error(("can't set value for %s, it's readonly."):format(_NSInfo[self].Name), 2)
		end

		_MetaNS.__add = function(v1, v2)
			local fenv = getfenv(2)

			local _type1 = BuildType(v1, fenv)
			local _type2 = BuildType(v2, fenv)

			return _type1 + _type2
		end
		
		_MetaNS.__tostring = function(self)			
			local info = _NSInfo[self]

			if info.Type == TYPE_ENUM then
			
				local str = info.Name.." { "
				
				for i in pairs(info.Enum) do
					str = str..i..", "
				end
				str = str.."}"
				
				return str
			else
				return info.Name
			end
		end

		_MetaNS.__metatable = TYPE_NAMESPACE
	end

	-- IsNameSpace
	function IsNameSpace(ns)
		return ns and type(ns) == "userdata" and getmetatable(ns) == TYPE_NAMESPACE or false
	end

	-- BuildNameSpace
	function BuildNameSpace(ns, namelist)
		if type(namelist) ~= "string" or (ns ~= nil and not IsNameSpace(ns)) then
			return
		end

		if namelist:find("%.%s*%.") then
			error("the namespace 's name can't have empty string between dots.", 2)
		end

		local cls = ns
		local info = cls and _NSInfo[cls]

		for name in namelist:gmatch("[^%.]+") do
			name = name:match("[_%w]+")

			if not name or name == "" then
				error("the namespace's name must be composed with number, string or '_'.", 2)
			end

			if not info then
				cls = newproxy(_NameSpace)
			elseif info.Type == nil or info.Type == TYPE_CLASS or info.Type == TYPE_STRUCT then
				info.SubNS = info.SubNS or {}
				info.SubNS[name] = info.SubNS[name] or newproxy(_NameSpace)

				cls = info.SubNS[name]
			else
				error(("can't add item to a %s."):format(tostring(info.Type)), 2)
			end

			info = _NSInfo[cls]
			info.Name = name
		end

		if cls == ns then
			return
		end

		return cls
	end

	-- GetNameSpace
	function GetNameSpace(ns, namelist)
		if type(namelist) ~= "string" or not IsNameSpace(ns) then
			return
		end

		if namelist:find("%.%s*%.") then
			error("the namespace 's name can't have empty string between dots.", 2)
		end

		local cls = ns

		for name in namelist:gmatch("[^%.]+") do
			name = name:match("[_%w]+")

			if not name or name == "" then
				error("the namespace's name must be composed with number, string or '_'.", 2)
			end

			cls = cls[name]
			if not cls then
				return
			end
		end

		if cls == ns then
			return
		end

		return cls
	end

	-- GetDefaultNameSpace
	function GetDefaultNameSpace()
		return _NameSpace
	end

	-- SetNameSpace
	function SetNameSpace4Env(env, name)
		if type(env) ~= "table" then
			return
		end

		if type(name) == "string" then
			local ns = BuildNameSpace(GetDefaultNameSpace(), name)

			if ns then
				rawset(env, "__IGAS_NameSpace", ns)
			else
				rawset(env, "__IGAS_NameSpace", nil)
			end
		elseif IsNameSpace(name) then
			rawset(env, "__IGAS_NameSpace", name)
		else
			rawset(env, "__IGAS_NameSpace", nil)
		end
	end

	-- GetEnvNameSpace
	function GetNameSpace4Env(env)
		local ns = type(env) == "table" and env.__IGAS_NameSpace
		if IsNameSpace(ns) then
			return ns
		end
	end

	------------------------------------
	--- Set the default namespace for the current environment, the class defined in this environment will be stored in this namespace
	-- @name namespace
	-- @class function
	-- @param name the namespace's name list, using "." to split.
	-- @usage namespace "Widget"
	------------------------------------
	function namespace(name)
		if name ~= nil and type(name) ~= "string" and not IsNameSpace(name) then
			error([[Usage: namespace "namespace"]], 2)
		end

		local fenv = getfenv(2)

		SetNameSpace4Env(fenv, name)
	end
end

------------------------------------------------------
-- Type
------------------------------------------------------
do
	_Meta4Type = _Meta4Type or {}
	do
		_Meta4Type.__add = function(v1, v2)
			local fenv = getfenv(2)

			local _type1 = BuildType(v1, fenv)
			local _type2 = BuildType(v2, fenv)

			if _type1 and _type2 then
				local _type = setmetatable({}, _Meta4Type)

				_type.AllowNil = _type1.AllowNil or _type2.AllowNil

				local tmp = {}

				for _, ns in ipairs(_type1) do
					tinsert(_type, ns)
					tmp[ns] = true
				end
				for _, ns in ipairs(_type2) do
					if not tmp[ns] then
						tinsert(_type, ns)
					end
				end

				tmp = nil

				return _type
			else
				return _type1 or _type2
			end
		end

		_Meta4Type.__index = {
			Validate = function(self, value)
				if value == nil and self.AllowNil then
					return value
				end

				local flag, msg, info, new

				for _, ns in ipairs(self) do
					info = _NSInfo[ns]

					if info.Type == TYPE_CLASS then
						-- Check if the value is an instance of this class
						if type(value) == "table" and getmetatable(value) and getmetatable(value).__class and IsChildClass(ns, getmetatable(value).__class) then
							return value
						else
							msg = msg or ("%s must be an instance of class '%s'."):format("%s", info.Name)
						end
					elseif info.Type == TYPE_ENUM then
						-- Check if the value is an enumeration value of this enum
						if type(value) == "string" and info.Enum[value:upper()] then
							return info.Enum[value:upper()]
						else
							msg = msg or ("%s must be a value of enum %s."):format("%s", tostring(ns))
						end
					elseif info.Type == TYPE_STRUCT and info.Validate then
						-- Check if the value is an enumeration value of this structure
						flag, new = pcall(info.Validate, value)

						if flag then
							return new
						elseif not msg then
							new = strtrim(new:match(":%d+:(.*)$") or new)

							if new:find("%%s([_%w]+)") then
								msg = new:gsub("%%s", "%%s"..self.Name..".")
							else
								msg = new:gsub("%%s", "%%s"..self.Name)
							end
						end
					end
				end

				assert(not msg, msg)

				return value
			end,
		}
	end

	function BuildType(ns, env, name)
		local allowNil = false

		if ns == nil then
			allowNil = true
		elseif type(ns) == "string" and env then
			if strfind(ns, "?$") then
				allowNil = true
			end
			ns = ns:match("^[_%w]+")
			ns = env[ns]
		elseif type(ns) == "table" and getmetatable(ns) == _Meta4Type then
			if name then
				ns.Name = name
			end
			return ns
		end

		if ns == nil or IsNameSpace(ns) then
			local _type = setmetatable({}, _Meta4Type)

			_type.AllowNil = allowNil or nil
			tinsert(_type, ns)

			if name then
				_type.Name = name
			end
			return _type
		else
			error("not supported type.", 2)
		end
	end
end

------------------------------------------------------
-- Class
------------------------------------------------------
do
	_ClsEnv2Info = _ClsEnv2Info or setmetatable({}, {__mode = "kv",})
	_ClsCache2Info = _ClsCache2Info  or setmetatable({}, {__mode = "kv",})
	_ClsMeta2Info = _ClsMeta2Info or  setmetatable({}, {__mode = "kv",})

	_KeyWord4ClsEnv = _KeyWord4ClsEnv or {}

	_KeyMeta = {
		__add = true,	-- a + b
		__sub = true,	-- a - b
		__mul = true,	-- a * b
		__div = true,	-- a / b
		__mod = true,	-- a % b
		__pow = true,	-- a ^ b
		__unm = true,	-- - a
		__concat = true,	-- a..b
		__len = true,	-- #a
		__eq = true,		-- a == b
		__lt = true,		-- a < b
		__le = true,		-- a <= b
		__index = false,	-- return a[b]
		__newindex = false,	-- a[b] = v
		__call = true,	-- a()
		__gc = false,		-- dispose a
		__tostring = true,	-- tostring(a)
		__exist = true,		-- ClassName(...)	-- return object if existed
	}

	-- metatable for class's env
	_MetaClsEnv = _MetaClsEnv or {}
	do
		_MetaClsEnv.__index = function(self, key)
			local info = _ClsEnv2Info[self]

			-- Check owner
			if key == info.Name then
				return info.Owner
			end

			-- Check keywords
			if _KeyWord4ClsEnv[key] then
				return _KeyWord4ClsEnv[key]
			end

			-- Check namespace
			if info.NameSpace then
				if key == _NSInfo[info.NameSpace].Name then
					self[key] = info.NameSpace
					return rawget(self, key)
				elseif info.NameSpace[key] then
					self[key] = info.NameSpace[key]
					return rawget(self, key)
				end
			end

			-- Check imports
			if info.Import4Env then
				for _, ns in ipairs(info.Import4Env) do
					if key == _NSInfo[ns].Name then
						self[key] = ns
						return rawget(self, key)
					elseif ns[key] then
						self[key] = ns[key]
						return rawget(self, key)
					end
				end
			end

			-- Check Base
			if info.BaseEnv then
				return info.BaseEnv[key]
			end
		end

		_MetaClsEnv.__newindex = function(self, key, value)
			local info = _ClsEnv2Info[self]

			if _KeyWord4ClsEnv[key] then
				error(("'%s' is a keyword."):format(key), 2)
			end

			if key == info.Name then
				if type(value) == "function" then
					rawset(info, "Constructor", value)
					return
				else
					error(("'%s' must be a function as constructor."):format(key), 2)
				end
			end

			if _KeyMeta[key] ~= nil then
				if type(value) == "function" then
					local rMeta = _KeyMeta[key] and key or "_"..key
					SetMetaFunc(rMeta, info.ChildClass, info.MetaTable[rMeta], value)
					info.MetaTable[rMeta] = value
					return
				else
					error(("'%s' must be a function."):format(key), 2)
				end
			end

			rawset(self, key, value)
		end
	end

	-- metatable for class cache
	_MetaClsCache4Script = _MetaClsCache4Script or {}
	do
		_BaseScripts = _BaseScripts or {
			OnScriptHandlerChanged = true,
		}

		_MetaClsCache4Script.__index = function(self, key)
			local info = _ClsCache2Info[self]

			if info.Script[key] ~= nil then
				rawset(self, key, info.Script[key])
				return rawget(self, key)
			end

			if info.SuperClass then
				if _NSInfo[info.SuperClass].Cache4Script[key] ~= nil then
					rawset(self, key, _NSInfo[info.SuperClass].Cache4Script[key])
					return rawget(self, key)
				end
			elseif _BaseScripts[key] ~= nil then
				rawset(self, key,_BaseScripts[key])
				return rawget(self, key)
			end
		end
	end

	_MetaClsCache4Prop = _MetaClsCache4Prop or {}
	do
		_MetaClsCache4Prop.__index = function(self, key)
			local info = _ClsCache2Info[self]

			if info.Property[key] then
				rawset(self, key, info.Property[key])
				return rawget(self, key)
			end

			if info.SuperClass and _NSInfo[info.SuperClass].Cache4Property[key] then
				rawset(self, key, _NSInfo[info.SuperClass].Cache4Property[key])
				return rawget(self, key)
			end
		end
	end

	_MetaClsCache4Method = _MetaClsCache4Method or {}
	do
		_MetaClsCache4Method.__index = function(self, key)
			local info = _ClsCache2Info[self]

			if type(rawget(info.ClassEnv, key)) == "function" then
				rawset(self, key, rawget(info.ClassEnv, key))
				return rawget(self, key)
			end

			if info.SuperClass and _NSInfo[info.SuperClass].Cache4Method[key] ~= nil then
				rawset(self, key, _NSInfo[info.SuperClass].Cache4Method[key])
				return rawget(self, key)
			end
		end
	end

	-- metatable for ScriptHandler
	_MetaScriptHandler = _MetaScriptHandler or {}
	do
		_MetaScriptHandler.__index = {
			Add = function(self, func)
				if type(func) ~= "function" then
					error("Usage: obj.OnXXXX:Add(func)", 2)
				end

				for _, f in ipairs(self) do
					if f == func then
						return self
					end
				end

				tinsert(self, func)

				if self._Owner and self._Name then
					CallScriptWithoutCreate(self._Owner, "OnScriptHandlerChanged", self._Name)
				end

				return self
			end,
			Remove = function(self, func)
				local flag = false

				if type(func) ~= "function" then
					error("Usage: obj.OnXXXX:Remove(func)", 2)
				end

				for i, f in ipairs(self) do
					if f == func then
						tremove(self, i)
						flag = true
						break
					end
				end

				if flag and self._Owner and self._Name then
					CallScriptWithoutCreate(self._Owner, "OnScriptHandlerChanged", self._Name)
				end

				return self
			end,
			Clear = function(self)
				local flag = false

				for i = #self, 0, -1 do
					if not flag and self[i] then
						flag = true
					end
					tremove(self, i)
				end

				if flag and self._Owner and self._Name then
					CallScriptWithoutCreate(self._Owner, "OnScriptHandlerChanged", self._Name)
				end

				return self
			end,
			Empty = function(self)
				return #self == 0 and self[0] == nil
			end,
		}

		_MetaScriptHandler.__add = function(self, func)
			if type(func) ~= "function" then
				error("Usage: obj.OnXXXX = obj.OnXXXX + func", 2)
			end

			for _, f in ipairs(self) do
				if f == func then
					return self
				end
			end

			tinsert(self, func)

			if self._Owner and self._Name then
				CallScriptWithoutCreate(self._Owner, "OnScriptHandlerChanged", self._Name)
			end

			return self
		end

		_MetaScriptHandler.__sub = function(self, func)
			local flag = false

			if type(func) ~= "function" then
				error("Usage: obj.OnXXXX = obj.OnXXXX - func", 2)
			end

			for i, f in ipairs(self) do
				if f == func then
					tremove(self, i)
					flag = true
					break
				end
			end

			if flag and self._Owner and self._Name then
				CallScriptWithoutCreate(self._Owner, "OnScriptHandlerChanged", self._Name)
			end

			return self
		end

		_MetaScriptHandler.__call = function(self, obj, ...)
			if not obj then
				error("Usage: obj:OnXXXX(...).", 2)
			end

			local chk = false
			local ret = false

			for i = 1, #self do
				chk, ret = pcall(self[i], obj, ...)
				if not chk then
					return errorhandler(ret)
				end

				if not rawget(obj, "__Scripts") then
					-- means it's disposed
					ret = true
				end
				if ret then
					break
				end
			end

			if not ret and self[0] then
				chk, ret = pcall(self[0], obj, ...)
				if not chk then
					return errorhandler(ret)
				end
			end
		end
	end

	_MetaScripts = _MetaScripts or {}
	do
		_MetaScripts.__index = function(self, key)
			-- Check Script
			local cls = self._Owner and getmetatable(self._Owner) and getmetatable(self._Owner).__class

			if _NSInfo[cls].Cache4Script[key] == nil then
				return
			end

			-- Add Script Handler
			rawset(self, key, setmetatable({_Owner = self._Owner, _Name = key}, _MetaScriptHandler))
			return rawget(self, key)
		end

		_MetaScripts.__newindex = function(self, key, value)
		end
	end

	function IsChildClass(cls, child)
		if not cls or not child or not _NSInfo[cls] or _NSInfo[cls].Type ~= TYPE_CLASS or not _NSInfo[child] or _NSInfo[child].Type ~= TYPE_CLASS then
			return false
		end

		if cls == child then
			return true
		end

		local info = _NSInfo[child]

		while info and info.SuperClass and info.SuperClass ~= cls do
			info = _NSInfo[info.SuperClass]
		end

		if info and info.SuperClass == cls then
			return true
		end

		return false
	end

	function SetMetaFunc(meta, sub, pre, now)
		if sub and pre ~= now then
			for cls in pairs(sub) do
				local info = _NSInfo[cls]

				if info.MetaTable[meta] == pre then
					info.MetaTable[meta] = now

					SetMetaFunc(meta, info.ChildClass, pre, now)
				end
			end
		end
	end

	function ClearCache(cls)
		local info = _NSInfo[cls]

		wipe(info.Cache4Script)
		wipe(info.Cache4Property)
		wipe(info.Cache4Method)

		-- Clear child's cache
		if info.ChildClass then
			for subcls in pairs(info.ChildClass) do
				ClearCache(subcls)
			end
		end
	end

	function CheckProperty(prop, value)
		if prop and prop.Type then
			local ok, ret = pcall(prop.Type.Validate, prop.Type, value)

			if not ok then
				ret = strtrim(ret:match(":%d+:(.*)$") or ret)

				if ret:find("%%s") then
					ret = ret:gsub("%%s[_%w]*", prop.Name)
				end

				error(ret, 3)
			end

			return ret
		end
		
		return value
	end

	function CallScriptWithoutCreate(self, scriptName, ...)
		if rawget(self, "__Scripts") and rawget(self.__Scripts, scriptName) then
			return self[scriptName](self, ...)
		end
	end

	function Class2Obj(cls, ...)
		local ret, obj

		local info = _NSInfo[cls]
		local cache = info.Cache4Method

		-- Check if this class has __exist so no need to create again.
		if type(info.MetaTable.__exist) == "function" then
			ret = {pcall(info.MetaTable.__exist, cls, ...)}

			if not ret[1] then
				ret[2] = strtrim(ret[2]:match(":%d+:(.*)$") or ret[2])

				error(ret[2], 3)
			end

			if type(ret[2]) == "table" then
				obj = ret[2]
				local meta = getmetatable(obj)

				if meta and meta.__class then
					if meta ~= info.MetaTable then
						setmetatable(obj, info.MetaTable)
					end
					return obj
				end
			elseif ret[2] == false then
				-- if the __exist modify the args.

				-- Create new object
				local constructor = info.Constructor
				if type(constructor) ~= "function" then
					error(("no Constructor found for class %s."):format(info.Name), 3)
				end

				obj = constructor(unpack(ret, 3))

				if type(obj) == "table" then
					setmetatable(obj, info.MetaTable)
				end

				return obj
			end
		end

		-- Create new object
		local constructor = info.Constructor
		if type(constructor) ~= "function" then
			error(("no Constructor found for class %s."):format(info.Name), 3)
		end

		obj = constructor(...)

		if type(obj) == "table" then
			setmetatable(obj, info.MetaTable)
		end

		return obj
	end

	------------------------------------
	--- Create class in currect environment's namespace or default namespace
	-- @name class
	-- @class function
	-- @param name the class's name
	-- @usage class "Form"
	------------------------------------
	function class(name)
		if type(name) ~= "string" or name:find("%.") or not name:match("[_%w]+") then
			error([[Usage: class "classname"]], 2)
		end
		local fenv = getfenv(2)
		local ns = GetNameSpace4Env(fenv)

		-- Create class or get it
		local cls

		name = name:match("[_%w]+")

		if ns then
			cls = BuildNameSpace(ns, name)

			if _NSInfo[cls] then
				if _NSInfo[cls].Type and _NSInfo[cls].Type ~= TYPE_CLASS then
					error(("%s is existed as %s, not class."):format(name, tostring(_NSInfo[cls].Type)), 2)
				end

				if _NSInfo[cls].BaseEnv and _NSInfo[cls].BaseEnv ~= fenv then
					error(("%s is defined in another place, can't be defined here."):format(name), 2)
				end
			end
		else
			cls = fenv[name]

			if not(_NSInfo[cls] and _NSInfo[cls].BaseEnv == fenv and _NSInfo[cls].NameSpace == nil and _NSInfo[cls].Type == TYPE_CLASS) then
				cls = BuildNameSpace(nil, name)
			end
		end

		if not cls then
			error("no class is created.", 2)
		end

		-- save class to the environment
		rawset(fenv, name, cls)

		-- Build class
		info = _NSInfo[cls]
		info.Type = TYPE_CLASS
		info.NameSpace = ns
		info.BaseEnv = fenv
		info.Script = info.Script or {}
		info.Property = info.Property or {}

		info.ClassEnv = info.ClassEnv or setmetatable({}, _MetaClsEnv)
		_ClsEnv2Info[info.ClassEnv] = info

		-- Clear
		info.Constructor = nil
		wipe(info.Property)
		wipe(info.Script)
		for i, v in pairs(info.ClassEnv) do
			if type(v) == "function" then
				info.ClassEnv[i] = nil
			end
		end

		-- Set namespace
		SetNameSpace4Env(info.ClassEnv, cls)

		-- Cache
		info.Cache4Script = info.Cache4Script or setmetatable({}, _MetaClsCache4Script)
		info.Cache4Property = info.Cache4Property or setmetatable({}, _MetaClsCache4Prop)
		info.Cache4Method = info.Cache4Method or setmetatable({}, _MetaClsCache4Method)
		_ClsCache2Info[info.Cache4Script] = info
		_ClsCache2Info[info.Cache4Property] = info
		_ClsCache2Info[info.Cache4Method] = info

		ClearCache(cls)

		-- SuperClass
		local prevInfo = info.SuperClass and _NSInfo[info.SuperClass]

		if prevInfo and prevInfo.ChildClass then
			prevInfo.ChildClass[info.Owner] = nil
		end

		info.SuperClass = nil

		-- Import
		info.Import4Env = info.Import4Env or {}
		wipe(info.Import4Env)

		-- MetaTable
		info.MetaTable = info.MetaTable or {}
		do
			local MetaTable = info.MetaTable
			local rMeta

			-- Clear
			for meta, flag in pairs(_KeyMeta) do
				rMeta = flag and meta or "_"..meta
				SetMetaFunc(rMeta, info.ChildClass, MetaTable[rMeta], nil)
				MetaTable[rMeta] = nil
			end

			local ClassEnv = info.ClassEnv
			local Cache4Script = info.Cache4Script
			local Cache4Property = info.Cache4Property
			local Cache4Method = info.Cache4Method
			local ClassName = info.Name

			MetaTable.__class = cls

			MetaTable.__index = function(self, key)
				-- Property Get
				if Cache4Property[key] then
					if Cache4Property[key]["Get"] then
						return Cache4Property[key]["Get"](self)
					else
						error(("%s is write-only."):format(tostring(key)),2)
					end
				end

				-- Frame Func
				if type(key) == "string" and (not key:find("^_")) and Cache4Method and Cache4Method[key] then
					return Cache4Method[key]
				end

				-- Scripts
				if Cache4Script[key] ~= nil then
					if type(self.__Scripts) ~= "table" and getmetatable(self.__Scripts) ~= _MetaScripts then
						rawset(self, "__Scripts", setmetatable({_Owner = self}, _MetaScripts))
					end

					return self.__Scripts[key]
				end

				-- Custom index metametods
				if rawget(MetaTable, "___index") then
					if type(rawget(MetaTable, "___index")) == "table" then
						return rawget(MetaTable, "___index")[key]
					elseif type(rawget(MetaTable, "___index")) == "function" then
						return rawget(MetaTable, "___index")(self, key)
					end
				end
			end

			MetaTable.__newindex = function(self, key, value)
				-- Property Set
				if Cache4Property[key] then
					if Cache4Property[key]["Set"] then
						return Cache4Property[key]["Set"](self, CheckProperty(Cache4Property[key], value))
					else
						error(("%s is read-only."):format(tostring(key)),2)
					end
				end

				-- Scripts
				if Cache4Script[key] ~= nil then
					if type(self.__Scripts) ~= "table" and getmetatable(self.__Scripts) ~= _MetaScripts then
						rawset(self, "__Scripts", setmetatable({_Owner = self}, _MetaScripts))
					end

					if value == nil or type(value) == "function" then
						if Cache4Script[key] then
							if self.__Scripts[key][0] ~= value then
								rawset(self.__Scripts[key], 0, value)
								CallScriptWithoutCreate(self, "OnScriptHandlerChanged", key)
							end
						else
							error(("%s is not supported for class '%s'."):format(tostring(key), ClassName), 2)
						end
					elseif type(value) == "table" and getmetatable(value) == _MetaScriptHandler then
						if value == self.__Scripts[key] then
							return
						end

						for i = #self.__Scripts[key], 0, -1 do
							tremove(self.__Scripts[key], i)
						end

						for i =0, #value do
							self.__Scripts[key][i] = value[i]
						end

						CallScriptWithoutCreate(self, "OnScriptHandlerChanged", key)
					else
						error("can't set this value to a scipt handler.", 2)
					end

					return
				end

				-- Custom newindex metametods
				if type(rawget(MetaTable, "___newindex")) == "function" then
					return rawget(MetaTable, "___newindex")(self, key, value)
				end

				rawset(self,key,value)			-- Other key can be set as usual
			end
		end

		-- Set the environment to class's environment
		setfenv(2, info.ClassEnv)
	end

	------------------------------------
	--- Set the current class' super class
	-- @name inherit
	-- @class function
	-- @param name the namespace's name list, using "." to split.
	-- @usage inherit "System.Widget.Frame"
	------------------------------------
	function inherit(name)
		if name and type(name) ~= "string" and not IsNameSpace(name) then
			error([[Usage: inherit "namespace.classname"]], 2)
		end

		if type(name) == "string" and name:find("%.%s*%.") then
			error("the namespace 's name can't have empty string between dots.", 2)
		end

		local env = getfenv(2)

		if rawget(_ClsEnv2Info, env) == nil then
			error("can't using inherit here.", 2)
		end

		local info = _ClsEnv2Info[env]

		local prevInfo = info.SuperClass and _NSInfo[info.SuperClass]

		if prevInfo then
			error("this class already have a super class.", 2)
		end

		local superCls

		if type(name) == "string" then
			superCls = GetNameSpace(info.NameSpace, name)

			if not superCls then
				for subname in name:gmatch("[^%.]+") do
					subname = subname:match("[_%w]+")

					if not subname or subname == "" then
						error("the namespace's name must be composed with number, string or '_'.", 2)
					end

					if not superCls then
						superCls = info.BaseEnv[subname]
					else
						superCls = superCls[subname]
					end

					if not IsNameSpace(superCls) then
						error(("no class is found with the name : %s"):format(name), 2)
					end
				end
			end
		else
			superCls = name
		end

		local superInfo = _NSInfo[superCls]

		if not superInfo or superInfo.Type ~= TYPE_CLASS then
			error("Usage: inherit (class) : 'class' - class expected", 2)
		end

		if IsChildClass(info.Owner, superCls) then
			error(("%s is inherited from %s, can't be used as super class."):format(baseInfo.Name, info.Name), 2)
		end

		superInfo.ChildClass = superInfo.ChildClass or {}
		superInfo.ChildClass[info.Owner] = true
		info.SuperClass = superCls

		-- Copy Metatable
		local rMeta
		for meta, flag in pairs(_KeyMeta) do
			rMeta = flag and meta or "_"..meta

			if info.MetaTable[rMeta] == nil and superInfo.MetaTable[rMeta] then
				SetMetaFunc(rMeta, info.ChildClass, nil, superInfo.MetaTable[rMeta])
				info.MetaTable[rMeta] = superInfo.MetaTable[rMeta]
			end
		end
	end

	------------------------------------
	--- import classes from the given name's namespace to the current environment
	-- @name import
	-- @class function
	-- @param name the namespace's name list, using "." to split.
	-- @usage import "System.Widget"
	------------------------------------
	function import(name)
		if type(name) ~= "string" and not IsNameSpace(name) then
			error([[Usage: import "namespaceA.namespaceB"]], 2)
		end

		if type(name) == "string" and name:find("%.%s*%.") then
			error("the namespace 's name can't have empty string between dots.", 2)
		end

		local env = getfenv(2)

		local info = _ClsEnv2Info[env]

		if not info then
			error("can't use import here.", 2)
		end

		local ns

		if type(name) == "string" then
			ns = GetNameSpace(GetDefaultNameSpace(), name)
		elseif IsNameSpace(name) then
			ns = name
		end

		if not ns then
			error(("no namespace is found with name : %s"):format(name), 2)
		end

		info.Import4Env = info.Import4Env or {}

		for _, v in ipairs(info.Import4Env) do
			if v == ns then
				return
			end
		end

		tinsert(info.Import4Env, ns)
	end

	------------------------------------
	--- Add or remove a script for current class
	-- @name script
	-- @class function
	-- @param name the name of the script, if started with "-" means to remove this script
	-- @usage script "OnClick"
	-- @usage script "-OnClick"
	------------------------------------
	function script(name)
		if type(name) ~= "string" or name:find("^_") then
			error([[Usage: script "[-]scriptName"]], 2)
		end

		local env = getfenv(2)

		local info = _ClsEnv2Info[env]

		if not info then
			error("can't use script here.", 2)
		end

		local flag

		name, flag = name:gsub("^-", "")

		info.Script[name] = (flag == 0)
	end

	local function SetProperty2Cls(env, info, name, set)
		if type(set) ~= "table" then
			error([=[Usage: property "propertyName" {
				Get = function(self)
					-- return the property's value
				end,
				Set = function(self, value)
					-- Set the property's value
				end,
				Type = Type1 [+ Type2 [+ nil]],	-- set the property's type
			}]=], 2)
		end

		local prop = info.Property[name] or {}
		info.Property[name] = prop

		wipe(prop)

		prop.Name = name
		prop.Get = type(set.Get) == "function" and set.Get
		prop.Set = type(set.Set) == "function" and set.Set

		if set.Type then
			prop.Type = BuildType(set.Type, env, name)
		end
	end

	------------------------------------
	--- set a propert to the current class
	-- @name property
	-- @class function
	-- @param name the name of the property
	-- @usage property "Title" {
	--		Get = function(self)
	--			-- return the property's value
	--		end,
	--		Set = function(self, value)
	--			-- Set the property's value
	--		end,
	--		Type = "XXXX",	-- set the property's type
	--}
	------------------------------------
	function property(name)
		if type(name) ~= "string" or strtrim(name:match("[_%w]+")) == "" then
			error([=[Usage: property "propertyName" {
				Get = function(self)
					-- return the property's value
				end,
				Set = function(self, value)
					-- Set the property's value
				end,
				Type = Type1 [+ Type2 [+ nil]],	-- set the property's type
			}]=], 2)
		end

		name = name:match("[_%w]+")

		local env = getfenv(2)

		local info = _ClsEnv2Info[env]

		if not info then
			error("can't use property here.", 2)
		end

		return function(set)
			return SetProperty2Cls(env, info, name, set)
		end
	end

	------------------------------------
	--- End the class's definition and restore the environment
	-- @name class
	-- @class function
	-- @param name the name of the class
	-- @usage endclass "Form"
	------------------------------------
	function endclass(name)
		if type(name) ~= "string" or name:find("%.") then
			error([[Usage: endclass "classname"]], 2)
		end

		local env = getfenv(2)

		while rawget(_ClsEnv2Info, env) do
			local info = _ClsEnv2Info[env]

			if info.Name == name then
				setfenv(2, info.BaseEnv)
				return
			end

			env = info.BaseEnv
		end

		error(("no class is found with name: %s"):format(name), 2)
	end

	_KeyWord4ClsEnv.class = class
	_KeyWord4ClsEnv.inherit = inherit
	_KeyWord4ClsEnv.import = import
	_KeyWord4ClsEnv.script = script
	_KeyWord4ClsEnv.property = property
	_KeyWord4ClsEnv.endclass = endclass
end

------------------------------------------------------
-- Enum
------------------------------------------------------
do
	local function BuildEnum(info, set)
		if type(set) ~= "table" then
			error([[Usage: enum "enumName" {
				"enumValue1",
				"enumValue2",
			}]], 2)
		end

		info.Enum = info.Enum or {}

		wipe(info.Enum)

		for i, v in pairs(set) do
			if type(i) == "string" then
				info.Enum[i:upper()] = v
			elseif type(v) == "string" then
				info.Enum[v:upper()] = v
			end
		end
	end

	------------------------------------
	--- create a enumeration
	-- @name enum
	-- @class function
	-- @param name the name of the enum
	-- @usage enum "ButtonState" {
	--		"PUSHED",
	--		"NORMAL",
	--}
	------------------------------------
	function enum(name)
		if type(name) ~= "string" or name:find("%.") or strtrim(name:match("[_%w]+") or "") == "" then
			error([[Usage: enum "enumName" {
				"enumValue1",
				"enumValue2",
			}]], 2)
		end

		name = name:match("[_%w]+")

		local fenv = getfenv(2)
		local ns = GetNameSpace4Env(fenv)

		-- Create class or get it
		local enm

		if ns then
			enm = BuildNameSpace(ns, name)

			if _NSInfo[cls] then
				if _NSInfo[cls].Type and _NSInfo[cls].Type ~= TYPE_ENUM then
					error(("%s is existed as %s, not enumeration."):format(name, tostring(_NSInfo[cls].Type)), 2)
				end
			end
		else
			enm = fenv[name]

			if not(_NSInfo[enm] and _NSInfo[enm].Type == TYPE_ENUM) then
				enm = BuildNameSpace(nil, name)
			end
		end

		if not enm then
			error("no enumeration is created.", 2)
		end

		-- save class to the environment
		rawset(fenv, name, enm)

		-- Build enm
		local info = _NSInfo[enm]
		info.Type = TYPE_ENUM

		return function(set)
			return BuildEnum(info, set)
		end
	end
end

------------------------------------------------------
-- Struct
------------------------------------------------------
do
	_StructEnv2Info = _StructEnv2Info or setmetatable({}, {__mode = "kv",})

	_KeyWord4StrtEnv = _KeyWord4StrtEnv or {}

	-- metatable for class's env
	_MetaStrtEnv = _MetaStrtEnv or {}
	do
		local function BuildWrapValidate(func)
			return function (...)
				local ok, ret = pcall(func, ...)

				if ok then
					return ret
				else
					ret = strtrim(ret:match(":%d+:(.*)$") or ret)
					ret = ret:gsub("%%s%.", ""):gsub("%%s", "")

					error(ret, 2)
				end
			end
		end

		_MetaStrtEnv.__index = function(self, key)
			local info = _StructEnv2Info[self]

			-- Check owner
			if key == info.Name then
				return info.Owner
			end

			if key == "Validate" then
				return info.WrapValidate
			end

			-- Check keywords
			if _KeyWord4StrtEnv[key] then
				return _KeyWord4StrtEnv[key]
			end

			-- Check namespace
			if info.NameSpace then
				if key == _NSInfo[info.NameSpace].Name then
					rawset(self, key, info.NameSpace)
					return rawget(self, key)
				elseif info.NameSpace[key] then
					rawset(self, key, info.NameSpace[key])
					return rawget(self, key)
				end
			end

			-- Check imports
			if info.Import4Env then
				for _, ns in ipairs(info.Import4Env) do
					if key == _NSInfo[ns].Name then
						self[key] = ns
						return rawget(self, key)
					elseif ns[key] then
						self[key] = ns[key]
						return rawget(self, key)
					end
				end
			end

			-- Check Base
			if info.BaseEnv then
				return info.BaseEnv[key]
			end
		end

		_MetaStrtEnv.__newindex = function(self, key, value)
			local info = _StructEnv2Info[self]

			if _KeyWord4StrtEnv[key] then
				error(("the '%s' is a keyword."):format(key), 2)
			end

			if key == info.Name then
				if type(value) == "function" then
					rawset(info, "Constructor", value)
					return
				else
					error(("the '%s' must be a function as constructor."):format(key), 2)
				end
			end

			if key == "Validate" then
				if value == nil then
					rawset(info, "Validate", nil)
					rawset(info, "WrapValidate", nil)
					return
				elseif type(value) == "function" then
					rawset(info, "Validate", value)
					rawset(info, "WrapValidate", BuildWrapValidate(value))
					return
				else
					error(("the '%s' must be a function used for validation."):format(key), 2)
				end
			end

			if type(key) == "string" and type(value) == "string" or type(value) == "table" or IsNameSpace(value) then
				local ok, ret = pcall(BuildType, value, self, key)

				if ok then
					rawset(self, key, ret)
					return
				end
			end

			rawset(self, key, value)
		end
	end

	function Struct2Obj(strt, ...)
		local info = _NSInfo[strt]

		if type(info.Constructor) == "function" then
			local ok, ret = pcall(info.Constructor, ...)
			if ok then
				return ret
			else
				ret = strtrim(ret:match(":%d+:(.*)$") or ret)

				error(ret, 3)
			end
		else
			error(("struct '%s' is abstract."):format(info.Name), 3)
		end
	end

	------------------------------------
	--- create a structure
	-- @name struct
	-- @class function
	-- @param name the name of the enum
	-- @usage struct "ButtonState" {
	--
	--}
	------------------------------------
	function struct(name)
		if type(name) ~= "string" or name:find("%.") or not name:match("[_%w]+") then
			error([[Usage: struct "structname"]], 2)
		end
		local fenv = getfenv(2)
		local ns = GetNameSpace4Env(fenv)

		-- Create class or get it
		local strt

		name = name:match("[_%w]+")

		if ns then
			strt = BuildNameSpace(ns, name)

			if _NSInfo[strt] then
				if _NSInfo[strt].Type and _NSInfo[strt].Type ~= TYPE_STRUCT then
					error(("%s is existed as %s, not struct."):format(name, tostring(_NSInfo[strt].Type)), 2)
				end

				if _NSInfo[strt].BaseEnv and _NSInfo[strt].BaseEnv ~= fenv then
					error(("%s is defined in another place, can't be defined here."):format(name), 2)
				end
			end
		else
			strt = fenv[name]

			if not(_NSInfo[strt] and _NSInfo[strt].BaseEnv == fenv and _NSInfo[strt].NameSpace == nil and _NSInfo[strt].Type == TYPE_STRUCT) then
				strt = BuildNameSpace(nil, name)
			end
		end

		if not strt then
			error("no struct is created.", 2)
		end

		-- save class to the environment
		rawset(fenv, name, strt)

		-- Build class
		info = _NSInfo[strt]
		info.Type = TYPE_STRUCT
		info.NameSpace = ns
		info.BaseEnv = fenv

		info.StructEnv = info.StructEnv or setmetatable({}, _MetaStrtEnv)
		_StructEnv2Info[info.StructEnv] = info

		-- Clear
		info.Constructor = nil

		wipe(info.StructEnv)

		-- Set namespace
		SetNameSpace4Env(info.StructEnv, strt)

		-- Set the environment to class's environment
		setfenv(2, info.StructEnv)
	end

	------------------------------------
	--- import classes from the given name's namespace to the current environment
	-- @name import
	-- @class function
	-- @param name the namespace's name list, using "." to split.
	-- @usage import "System.Widget"
	------------------------------------
	function import(name)
		if type(name) ~= "string" and not IsNameSpace(name) then
			error([[Usage: import "namespaceA.namespaceB"]], 2)
		end

		if type(name) == "string" and name:find("%.%s*%.") then
			error("the namespace 's name can't have empty string between dots.", 2)
		end

		local env = getfenv(2)

		local info = _StructEnv2Info[env]

		if not info then
			error("can't use import here.", 2)
		end

		local ns

		if type(name) == "string" then
			ns = GetNameSpace(GetDefaultNameSpace(), name)
		elseif IsNameSpace(name) then
			ns = name
		end

		if not ns then
			error(("no namespace is found with name : %s"):format(name), 2)
		end

		info.Import4Env = info.Import4Env or {}

		for _, v in ipairs(info.Import4Env) do
			if v == ns then
				return
			end
		end

		tinsert(info.Import4Env, ns)
	end

	------------------------------------
	--- End the class's definition and restore the environment
	-- @name class
	-- @class function
	-- @param name the name of the class
	-- @usage endclass "Form"
	------------------------------------
	function endstruct(name)
		if type(name) ~= "string" or name:find("%.") then
			error([[Usage: endstruct "structname"]], 2)
		end

		local env = getfenv(2)

		while rawget(_StructEnv2Info, env) do
			local info = _StructEnv2Info[env]

			if info.Name == name then
				setfenv(2, info.BaseEnv)
				return
			end

			env = info.BaseEnv
		end

		error(("no struct is found with name: %s"):format(name), 2)
	end

	_KeyWord4StrtEnv.struct = struct
	_KeyWord4StrtEnv.import = import
	_KeyWord4StrtEnv.endstruct = endstruct
end

------------------------------------------------------
-- Global Settings
------------------------------------------------------
do
	--[[
	_G.class = class
	_G.enum = enum
	_G.namespace = namespace
	_G.struct = struct	--]]

	_Class_KeyWords.namespace = namespace
	_Class_KeyWords.class = class
	_Class_KeyWords.enum = enum
	_Class_KeyWords.struct = struct

	------------------------------------
	--- Get the information of the class
	-- @name IGAS:GetClassInfo
	-- @class function
	-- @param class the class that need to check
	-- @return name
	-- @return type
	-- @return namespace
	-- @return superclass
	-- @usage IGAS:GetClassInfo(Object)
	------------------------------------
	function IGAS:GetClassInfo(cls)
		local info = _NSInfo[cls]

		if info then
			return info.Name, info.Type, info.NameSpace, info.SuperClass
		end
	end

	------------------------------------
	--- Get the GetNameSpace for the name
	-- @name IGAS:GetNameSpace
	-- @class function
	-- @param name the namespace's name, split by "."
	-- @param [default] the base namespace
	-- @return namespace the namespace
	-- @usage IGAS:GetNameSpace("System")
	------------------------------------
	function IGAS:GetNameSpace(name, default)
		default = default or GetDefaultNameSpace()

		return GetNameSpace(default, name)
	end

	------------------------------------
	--- Check if the class has that script
	-- @name IGAS:HasScript(class, script)
	-- @class function
	-- @param class the class that need to check
	-- @param script the script name
	-- @return true if the class has this script, otherwise false, nil when no class is found.
	-- @usage IGAS:HasScript(Addon, "OnEvent")
	------------------------------------
	function IGAS:HasScript(cls, sc)
		local info = _NSInfo[cls]

		if info and info.Type == TYPE_CLASS then
			return info.Cache4Script[sc] or false
		end
	end
end